from collections.abc import Callable
from pathlib import Path
from typing import Any, overload

from liblaf.grapes.typing import PathLike

from ._decode import DecHook, PydanticModelValidateOptions
from ._serde import json, toml, yaml

readers: dict[str, Callable] = {}
readers[".json"] = json.load
readers[".toml"] = toml.load
readers[".yaml"] = yaml.load
readers[".yml"] = yaml.load


@overload
def load(
    path: PathLike,
    /,
    *,
    dec_hook: DecHook | None = ...,
    force_ext: str | None = None,
    pydantic: PydanticModelValidateOptions | None = None,
    strict: bool = True,
) -> Any: ...
@overload
def load[T](
    path: PathLike,
    /,
    *,
    dec_hook: DecHook | None = ...,
    force_ext: str | None = None,
    pydantic: PydanticModelValidateOptions | None = None,
    strict: bool = True,
    type: type[T],
) -> T: ...
@overload
def load[T](
    path: PathLike,
    /,
    *,
    dec_hook: DecHook | None = ...,
    force_ext: str | None = None,
    pydantic: PydanticModelValidateOptions | None = None,
    strict: bool = True,
    type: Any,
) -> Any: ...
def load(path: PathLike, /, *, force_ext: str | None = None, **kwargs) -> Any:
    path = Path(path)
    ext: str = force_ext or path.suffix
    return readers[ext](path, **kwargs)



import contextlib
from collections.abc import Generator, Mapping
from typing import Any, Self

import pydantic
import pydantic_settings


class MixinOverrides:
    @contextlib.contextmanager
    def overrides(self, **kwargs) -> Generator[Self]:
        original: dict[str, Any] = {}
        try:
            for k, v in kwargs.items():
                original[k] = getattr(self, k)
                setattr(self, k, v)
            yield self
        finally:
            for k, v in original.items():
                setattr(self, k, v)


class BaseModel(MixinOverrides, pydantic.BaseModel):
    model_config = pydantic.ConfigDict(
        validate_assignment=True, arbitrary_types_allowed=True, validate_default=True
    )

    def to_dict(self, **kwargs) -> Mapping[str, Any]:
        return self.model_dump(**kwargs)


class BaseConfig(MixinOverrides, pydantic_settings.BaseSettings):
    model_config = pydantic_settings.SettingsConfigDict(
        env_nested_delimiter="__",
        validate_assignment=True,
        arbitrary_types_allowed=True,
        validate_default=True,
    )



import inspect
import itertools
import logging
import types
from collections.abc import Iterable, Sequence

import attrs
import loguru
import loguru._logger
from loguru import logger


@attrs.define(slots=False)
class MixinLoguru:
    handlers: Sequence["loguru.HandlerConfig"] | None = attrs.field(default=None)
    levels: Sequence["loguru.LevelConfig"] | None = attrs.field(default=None)
    level: int | str | None = attrs.field(default=None)

    def add_level(
        self,
        name: str,
        no: int | None = None,
        color: str | None = None,
        icon: str | None = None,
    ) -> "loguru.Level":
        if name in self._core.levels and no == logger.level(name).no:
            no = None  # skip update severity no
        return logger.level(name=name, no=no, color=color, icon=icon)

    def configure_loguru(
        self,
        handlers: Sequence["loguru.HandlerConfig"] | None = None,
        levels: Sequence["loguru.LevelConfig"] | None = None,
        level: int | str | None = None,
    ) -> None:
        if handlers is None:
            handlers = self.handlers or []
        if levels is None:
            levels = self.levels or []
        if level is None:
            level = self.level

        if level is not None:
            for handler_config in handlers:
                if "level" not in handler_config:
                    handler_config["level"] = level

        self.handlers = handlers
        self.levels = levels
        self.level = level

        for level_config in levels:
            self.add_level(**level_config)
        logger.configure(handlers=handlers)
        self.setup_loguru_logging_intercept()

    def setup_loguru_logging_intercept(self, modules: Iterable[str] = ()) -> None:
        """Logs to loguru from Python logging module.

        References:
            1. [Entirely compatible with standard logging](https://loguru.readthedocs.io/en/stable/overview.html#entirely-compatible-with-standard-logging)
            2. [loguru-logging-intercept/loguru_logging_intercept.py at f358b75ef4162ea903bf7a3298c22b1be83110da · MatthewScholefield/loguru-logging-intercept](https://github.com/MatthewScholefield/loguru-logging-intercept/blob/f358b75ef4162ea903bf7a3298c22b1be83110da/loguru_logging_intercept.py)
        """
        for lvl in self._core.levels.values():
            logging.addLevelName(lvl.no, lvl.name)

        level: int = logging.NOTSET
        if isinstance(self.level, int):
            level = self.level
        elif isinstance(self.level, str):
            level = logger.level(self.level).no

        logging.basicConfig(handlers=[InterceptHandler()], level=level)
        for logger_name in itertools.chain(("",), modules):
            mod_logger: logging.Logger = logging.getLogger(logger_name)
            mod_logger.handlers = [InterceptHandler()]
            mod_logger.propagate = False

    @property
    def _core(self) -> loguru._logger.Core:
        return logger._core  # pyright: ignore[reportAttributeAccessIssue] # noqa: SLF001


class InterceptHandler(logging.Handler):
    """Logs to loguru from Python logging module.

    References:
        1. [Entirely compatible with standard logging](https://loguru.readthedocs.io/en/stable/overview.html#entirely-compatible-with-standard-logging)
        2. [loguru-logging-intercept/loguru_logging_intercept.py at f358b75ef4162ea903bf7a3298c22b1be83110da · MatthewScholefield/loguru-logging-intercept](https://github.com/MatthewScholefield/loguru-logging-intercept/blob/f358b75ef4162ea903bf7a3298c22b1be83110da/loguru_logging_intercept.py)
    """

    def emit(self, record: logging.LogRecord) -> None:
        if logger is None:  # logger has been cleaned up
            return

        # Get corresponding Loguru level if it exists.
        level: str | int
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message.
        frame: types.FrameType | None
        depth: int
        frame, depth = inspect.currentframe(), 0
        while frame and (depth == 0 or frame.f_code.co_filename == logging.__file__):
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(
            level, record.getMessage()
        )



from .profiles import Profile, ProfileName, factory


def init(profile: ProfileName | Profile = "default") -> None:
    if isinstance(profile, str):
        profile = factory(profile)
    profile.init()



from ._abc import RichSinkColumn
from ._elapsed import RichSinkColumnElapsed
from ._level import RichSinkColumnLevel
from ._location import RichSinkColumnLocation
from ._message import RichSinkColumnMessage

__all__ = [
    "RichSinkColumn",
    "RichSinkColumnElapsed",
    "RichSinkColumnLevel",
    "RichSinkColumnLocation",
    "RichSinkColumnMessage",
]



from typing import override

import attrs
import loguru
from rich.console import RenderableType
from rich.highlighter import Highlighter, ReprHighlighter
from rich.text import Text

from ._abc import RichSinkColumn


@attrs.define
class RichSinkColumnMessage(RichSinkColumn):
    highlighter: Highlighter = attrs.field(factory=ReprHighlighter)

    @override  # impl RichSinkColumn
    def render(self, record: "loguru.Record") -> RenderableType:
        if (rich := record["extra"].get("rich")) is not None:
            return rich
        message: RenderableType = record["message"].strip()
        if "\x1b" in message:
            return Text.from_ansi(message)
        if record["extra"].get("markup", True):
            message = Text.from_markup(message)
        if highlighter := record["extra"].get("highlighter", self.highlighter):
            message = highlighter(message)
        return message



from collections.abc import Callable
from typing import Any

from liblaf.grapes import functools as _ft
from liblaf.grapes import pretty
from liblaf.grapes.logging import depth_tracker

from ._base import BaseTimer


def timed_callable[C: Callable](func: C, timer: BaseTimer) -> C:
    if timer.name is None:
        timer.name = pretty.pretty_func(func)

    @_ft.decorator
    @depth_tracker
    def wrapper(wrapped: C, _instance: Any, args: tuple, kwargs: dict[str, Any]) -> Any:
        timer.start()
        try:
            return wrapped(*args, **kwargs)
        finally:
            timer.stop()

    proxy: C = wrapper(func)
    proxy._self_timer = timer  # pyright: ignore[reportFunctionMemberAccess] # noqa: SLF001
    return proxy



from ._rich import (
    RichSink,
    RichSinkColumn,
    RichSinkColumnElapsed,
    RichSinkColumnLevel,
    RichSinkColumnLocation,
    RichSinkColumnMessage,
    RichTracebackConfig,
    default_columns,
    default_console,
    default_suppress,
)

__all__ = [
    "RichSink",
    "RichSinkColumn",
    "RichSinkColumnElapsed",
    "RichSinkColumnLevel",
    "RichSinkColumnLocation",
    "RichSinkColumnMessage",
    "RichTracebackConfig",
    "default_columns",
    "default_console",
    "default_suppress",
]



import functools
from typing import IO

import rich
from rich.console import Console
from rich.style import Style
from rich.theme import Theme

from liblaf.grapes import env
from liblaf.grapes.typing import clone_param_spec


def default_theme() -> Theme:
    """.

    References:
        1. <https://github.com/Delgan/loguru/blob/master/loguru/_defaults.py>
    """
    return Theme(
        {
            "logging.level.notset": Style(dim=True),
            "logging.level.trace": Style(color="cyan", bold=True),
            "logging.level.debug": Style(color="blue", bold=True),
            "logging.level.icecream": Style(color="magenta", bold=True),
            "logging.level.info": Style(bold=True),
            "logging.level.success": Style(color="green", bold=True),
            "logging.level.warning": Style(color="yellow", bold=True),
            "logging.level.error": Style(color="red", bold=True),
            "logging.level.critical": Style(color="red", bold=True, reverse=True),
        },
        inherit=True,
    )


@clone_param_spec(Console)
@functools.cache
def get_console(**kwargs) -> Console:
    if kwargs.get("theme") is None:
        kwargs["theme"] = default_theme()
    file: IO[str] | None = kwargs.get("file")
    if file is None and env.in_ci():
        kwargs.setdefault("width", 128)
    if not kwargs.get("stderr", False) and file is None:
        rich.reconfigure(**kwargs)
        return rich.get_console()
    return Console(**kwargs)



import contextlib
import functools
from collections.abc import Callable, Generator, Iterable
from typing import Literal, overload

import joblib
from rich.progress import Progress as RichProgress
from rich.progress import TaskID

from ._progress import Progress, len_safe


@overload
def parallel[T](
    fn: Callable[..., T],
    *iterables: Iterable,
    description: str = "Working...",
    progress: Progress | Literal[False] | None = None,
    return_as: Literal["list"] = "list",
    total: int | None = None,
) -> list[T]: ...
@overload
def parallel[T](
    fn: Callable[..., T],
    *iterables: Iterable,
    description: str = "Working...",
    progress: Progress | Literal[False] | None = None,
    return_as: Literal["generator", "generator_unordered"],
    total: int | None = None,
) -> Generator[T]: ...
def parallel[T](
    fn: Callable[..., T],
    *iterables: Iterable,
    description: str = "Working...",
    progress: Progress | Literal[False] | None = None,
    return_as: Literal["list", "generator", "generator_unordered"] = "list",
    total: int | None = None,
) -> list[T] | Generator[T]:
    for iterable in iterables:
        if total is not None:
            break
        total = len_safe(iterable)

    parallel = joblib.Parallel(return_as=return_as)

    task_id: TaskID | None = None
    if progress is None:
        progress = Progress()
    progress: Progress | contextlib.nullcontext
    if progress:
        task_id = progress.add_task(description=description, total=total)
    else:
        progress = contextlib.nullcontext()
    parallel.print_progress = functools.partial(
        print_progress, self=parallel, progress=progress, task_id=task_id
    )

    jobs = map(joblib.delayed(fn), *iterables)
    match return_as:
        case "list":
            return as_list(parallel, jobs, progress=progress)
        case "generator" | "generator_unordered":
            return as_generator(parallel, jobs, progress=progress)


def print_progress(
    self: joblib.Parallel,
    progress: RichProgress | contextlib.nullcontext,
    task_id: TaskID | None,
) -> None:
    if isinstance(progress, RichProgress):
        assert task_id is not None
        progress.update(task_id, total=self.n_tasks, completed=self.n_completed_tasks)


def as_list(
    parallel: joblib.Parallel,
    jobs: Iterable,
    *,
    progress: contextlib.AbstractContextManager,
) -> list:
    with progress, parallel:
        return parallel(jobs)  # pyright: ignore[reportReturnType]


def as_generator(
    parallel: joblib.Parallel,
    jobs: Iterable,
    *,
    progress: contextlib.AbstractContextManager,
) -> Generator:
    with progress, parallel:
        yield from parallel(jobs)



import sys
import types
from typing import Any

import attrs
from loguru import logger


@attrs.define(slots=False)
class MixinExceptHook:
    excepthook_level: int | str = attrs.field(default="CRITICAL")
    excepthook_message: Any = attrs.field(default="")

    def configure_excepthook(
        self, level: int | str | None = None, message: Any = None
    ) -> None:
        if level is not None:
            self.excepthook_level = level
        if message is not None:
            self.excepthook_message = message
        sys.excepthook = self.excepthook

    def excepthook(
        self,
        exc_type: type[BaseException],
        exc_value: BaseException,
        exc_traceback: types.TracebackType | None,
        /,
    ) -> None:
        logger.opt(exception=(exc_type, exc_value, exc_traceback)).log(
            self.excepthook_level, self.excepthook_message
        )



import contextlib
import types
from collections.abc import Callable, Iterable
from typing import Any, Self, overload, override

import attrs

from liblaf.grapes.logging import depth_tracker

from ._base import BaseTimer
from ._callable import timed_callable
from ._iterable import TimedIterable


@attrs.define
class Timer(BaseTimer, contextlib.AbstractContextManager):
    @override  # contextlib.AbstractContextManager
    @depth_tracker
    def __enter__(self) -> Self:
        self.start()
        return self

    @override  # contextlib.AbstractContextManager
    @depth_tracker
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
        /,
    ) -> None:
        self.stop()

    @overload
    def __call__[C: Callable](self, func: C, /) -> C: ...
    @overload
    def __call__[I: Iterable](self, iterable: I, /) -> I: ...
    def __call__(self, func_or_iterable: Callable | Iterable, /) -> Any:
        if callable(func_or_iterable):
            return timed_callable(func_or_iterable, self)
        return TimedIterable(func_or_iterable, self)



import contextlib
from collections.abc import Generator
from typing import Any, Self

import pydantic
import pydantic_settings as ps


class OverridesMixin:
    @contextlib.contextmanager
    def overrides(self, **kwargs) -> Generator[Self]:
        backup: dict[str, Any] = {}
        for k, v in kwargs.items():
            backup[k] = getattr(self, k)
            setattr(self, k, v)
        try:
            yield self
        finally:
            for k, v in backup.items():
                setattr(self, k, v)


class BaseConfig(OverridesMixin, ps.BaseSettings):
    model_config = ps.SettingsConfigDict(
        validate_assignment=True, arbitrary_types_allowed=True, validate_default=True
    )


class BaseModel(OverridesMixin, pydantic.BaseModel):
    model_config = pydantic.ConfigDict(
        validate_assignment=True, arbitrary_types_allowed=True, validate_default=True
    )



import math
import statistics
from collections.abc import Sequence

import autoregistry

from liblaf.grapes import human

type StatisticName = str


STATISTICS_REGISTRY = autoregistry.Registry(prefix="_compute_")


STATISTICS_REGISTRY["max"] = max
STATISTICS_REGISTRY["mean"] = statistics.mean
STATISTICS_REGISTRY["median"] = statistics.median
STATISTICS_REGISTRY["min"] = min
STATISTICS_REGISTRY["stdev"] = statistics.stdev
STATISTICS_REGISTRY["total"] = sum


def compute_statistic(series: Sequence[float], stat_name: StatisticName) -> float:
    try:
        return STATISTICS_REGISTRY[stat_name](series)
    except (ValueError, statistics.StatisticsError):
        return math.nan


def pretty_statistic(series: Sequence[float], stat_name: StatisticName) -> str:
    if stat_name == "mean+stdev":
        mean: float = compute_statistic(series, "mean")
        stdev: float = compute_statistic(series, "stdev")
        return human.human_duration_with_stdev(mean, stdev)
    value: float = compute_statistic(series, stat_name)
    return human.human_duration(value)



"""This module provides functions for converting numerical values into human-readable strings."""

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach_stub(__name__, __file__)
del lazy



import math
from collections.abc import Callable


def _mm_ss(sec: float) -> str:
    minute: float
    sec: float
    minute, sec = divmod(sec, 60)
    return f"{int(minute):02}:{int(sec):02}"


def _hh_mm_ss(sec: float) -> str:
    hour: float
    sec: float
    hour, sec = divmod(sec, 3600)
    return f"{int(hour):02}:{_mm_ss(sec)}"


def _d_hh_mm_ss(sec: float) -> str:
    day: float
    sec: float
    day, sec = divmod(sec, 86400)
    return f"{int(day)}d,{_hh_mm_ss(sec)}"


DEFAULT_TEMPLATES: list[tuple[float, Callable[[float], str]]] = [
    (1e-9, lambda sec: f"{sec * 1e9:#.2f} ns"),
    (1e-6, lambda sec: f"{sec * 1e9:#.3g} ns"),
    (1e-3, lambda sec: f"{sec * 1e6:#.3g} µs"),
    (1.0, lambda sec: f"{sec * 1e3:#.3g} ms"),
    (60, lambda sec: f"{sec:#.3g} s"),
    (3600, _mm_ss),
    (360000, _hh_mm_ss),
    (math.inf, _d_hh_mm_ss),
]


def human_duration(sec: float) -> str:
    """.

    Examples:
        >>> human_duration(math.nan)
        '?? s'
        >>> human_duration(1e-12)
        '0.00 ns'
        >>> human_duration(1e-11)
        '0.01 ns'
        >>> human_duration(1e-10)
        '0.10 ns'
        >>> human_duration(1e-9)
        '1.00 ns'
        >>> human_duration(1e-8)
        '10.0 ns'
        >>> human_duration(1e-7)
        '100. ns'
        >>> human_duration(1e-6)
        '1.00 µs'
        >>> human_duration(1e-5)
        '10.0 µs'
        >>> human_duration(1e-4)
        '100. µs'
        >>> human_duration(1e-3)
        '1.00 ms'
        >>> human_duration(1e-2)
        '10.0 ms'
        >>> human_duration(1e-1)
        '100. ms'
        >>> human_duration(1.0)
        '1.00 s'
        >>> human_duration(1e1)
        '10.0 s'
        >>> human_duration(1e2)
        '01:40'
        >>> human_duration(1e3)
        '16:40'
        >>> human_duration(1e4)
        '02:46:40'
        >>> human_duration(1e5)
        '27:46:40'
        >>> human_duration(1e6)
        '11d,13:46:40'
    """
    if not math.isfinite(sec):
        return "?? s"
    for threshold, template in DEFAULT_TEMPLATES:
        if sec < threshold:
            return template(sec)
    raise NotImplementedError


def human_duration_with_stdev(mean: float, stdev: float) -> str:
    if not math.isfinite(stdev):
        return human_duration(mean)
    return f"{human_duration(mean)} ± {human_duration(stdev)}"



from collections.abc import Callable
from typing import Any, Protocol, overload

import wrapt

from liblaf.grapes.sentinel import MISSING


class Decorator(Protocol):
    def __call__[T](self, wrapped: T, /) -> T: ...


class Wrapper(Protocol):
    def __call__(
        self, wrapped: Any, instance: Any, args: tuple, kwargs: dict[str, Any], /
    ) -> Any: ...


@overload
def decorator(
    wrapper: Wrapper,
    enabled: bool | Callable[[], None] | None = None,  # noqa: FBT001
    adapter: Any = None,
    proxy: Callable = ...,
) -> Decorator: ...
@overload
def decorator(
    wrapper: None = None,
    enabled: bool | Callable[[], None] | None = None,  # noqa: FBT001
    adapter: Any = None,
    proxy: Callable = ...,
) -> Callable[[Wrapper], Decorator]: ...
def decorator(*args, **kwargs) -> Any:
    return wrapt.decorator(*args, **kwargs)


def unbind[T](o: T) -> T:
    return getattr(o, "_self_parent", o)


@overload
def unbind_getattr(o: object, name: str, /) -> Any: ...
@overload
def unbind_getattr[T](o: object, name: str, default: T, /) -> Any | T: ...
def unbind_getattr(o: object, name: str, default: Any = MISSING, /) -> Any:
    try:
        return getattr(o, name)
    except AttributeError:
        parent: Any = getattr(o, "_self_parent", MISSING)
        if parent is MISSING:
            # `o` is not a wrapt.BoundFunctionWrapper
            # further inspection is not applicable
            if default is MISSING:
                raise
            return default
        # `o` is a wrapt.BoundFunctionWrapper
        # `parent` is a wrapt.FunctionWrapper
        # further inspection is applicable
        attr: Any = getattr(parent, name, MISSING)
        if attr is not MISSING:
            return attr
        if default is MISSING:
            raise
        return default



from pathlib import Path

import platformdirs
import pydantic
import pydantic_settings as ps

from ._base import BaseConfig


def _default_location() -> Path:
    return platformdirs.user_cache_path(appname="joblib")


class ConfigJoblibMemory(BaseConfig):
    model_config = ps.SettingsConfigDict(env_prefix="JOBLIB_MEMORY_")
    location: Path = pydantic.Field(default_factory=_default_location)
    bytes_limit: int | str | None = "4G"


class ConfigJoblib(BaseConfig):
    memory: ConfigJoblibMemory = pydantic.Field(default_factory=ConfigJoblibMemory)



import pydantic
import pydantic_settings

from ._base import BaseConfig
from ._joblib import ConfigJoblib
from ._logging import ConfigLogging
from ._pretty import ConfigPretty


class Config(BaseConfig):
    model_config = pydantic_settings.SettingsConfigDict(env_prefix="LIBLAF_GRAPES_")

    joblib: ConfigJoblib = pydantic.Field(default_factory=ConfigJoblib)
    logging: ConfigLogging = pydantic.Field(default_factory=ConfigLogging)
    pretty: ConfigPretty = pydantic.Field(default_factory=ConfigPretty)


config = Config()



from ._count import human_count
from ._duration import human_duration, human_duration_with_stdev
from ._throughout import human_throughput

__all__ = [
    "human_count",
    "human_duration",
    "human_duration_with_stdev",
    "human_throughput",
]



import pydantic

from ._base import BaseConfig
from ._joblib import ConfigJoblib
from ._pretty import ConfigPretty


class Config(BaseConfig):
    joblib: ConfigJoblib = pydantic.Field(default_factory=ConfigJoblib)
    pretty: ConfigPretty = pydantic.Field(default_factory=ConfigPretty)


config = Config()



from collections.abc import Sequence
from typing import Unpack

import loguru
from rich.console import Console

from liblaf.grapes.logging.filters import make_filter
from liblaf.grapes.logging.sink import RichSink, RichSinkColumn, default_columns


def rich_handler(
    columns: Sequence[RichSinkColumn] | None = None,
    console: Console | None = None,
    *,
    enable_link: bool = True,
    **kwargs: Unpack["loguru.BasicHandlerConfig"],
) -> "loguru.BasicHandlerConfig":
    if columns is None:
        columns = default_columns(enable_link=enable_link)
    kwargs["sink"] = RichSink(console=console, columns=columns)
    kwargs["format"] = ""
    kwargs["filter"] = make_filter(kwargs.get("filter"))
    return kwargs



import sys

import attrs
from loguru import logger


@attrs.define(slots=False)
class MixinUnraisableHook:
    unraisablehook_level: int | str = attrs.field(default="ERROR")

    def configure_unraisablehook(self, level: int | str | None = None) -> None:
        if level is not None:
            self.unraisablehook_level = level
        sys.unraisablehook = self.unraisablehook

    def unraisablehook(self, args: "sys.UnraisableHookArgs", /) -> None:
        if logger is None:  # logger has been cleaned up
            return
        logger.opt(exception=(args.exc_type, args.exc_value, args.exc_traceback)).log(
            self.unraisablehook_level,
            "{err_msg}: {object!r}",
            err_msg=args.err_msg or "Exception ignored in",
            object=args.object,
        )



import datetime
import functools
from collections.abc import Callable, Iterable, Mapping
from typing import Any, Literal, Protocol, TypedDict, overload

import cytoolz as toolz
import joblib

from liblaf.grapes.conf import config

from ._wrapt import decorator


class MemorizedFunc[**P, T](Protocol):
    @property
    def _self_memory(self) -> joblib.Memory: ...
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T: ...


class Metadata(TypedDict): ...


@overload
def cache[**P, T](
    func: Callable[P, T],
    /,
    *,
    memory: joblib.Memory | None = ...,
    # memory.cache() params
    ignore: list[str] | None = ...,
    verbose: int | None = ...,
    mmap_mode: Literal["r+", "r", "w+", "c"] | None = ...,
    cache_validation_callback: Callable[[Metadata], bool] | None = ...,
    # memory.reduce_size() params
    bytes_limit: int | str | None = ...,
    items_limit: int | None = ...,
    age_limit: datetime.timedelta | None = ...,
) -> MemorizedFunc[P, T]: ...
@overload
def cache[**P, T](
    *,
    memory: joblib.Memory | None = None,
    # memory.cache() params
    ignore: list[str] | None = ...,
    verbose: int | None = ...,
    mmap_mode: Literal["r+", "r", "w+", "c"] | None = ...,
    cache_validation_callback: Callable[[Metadata], bool] | None = ...,
    # memory.reduce_size() params
    bytes_limit: int | str | None = ...,
    items_limit: int | None = ...,
    age_limit: datetime.timedelta | None = ...,
) -> Callable[[Callable[P, T]], MemorizedFunc[P, T]]: ...
def cache(func: Callable | None = None, /, **kwargs: Any) -> Any:
    if func is None:
        return functools.partial(cache, **kwargs)
    memory: joblib.Memory | None = kwargs.pop("memory", None)
    if memory is None:
        memory = new_memory()
    cache_kwargs: dict[str, Any] = _filter_keys(
        kwargs, ("ignore", "verbose", "mmap_mode", "cache_validation_callback")
    )
    reduce_size_kwargs: dict[str, Any] = _filter_keys(
        kwargs, ("bytes_limit", "items_limit", "age_limit")
    )
    reduce_size_kwargs.setdefault("bytes_limit", config.joblib.memory.bytes_limit)

    @decorator
    def wrapper(
        wrapped: Callable, _instance: Any, args: tuple, kwargs: dict[str, Any]
    ) -> Any:
        result: Any = wrapped(*args, **kwargs)
        memory.reduce_size(**reduce_size_kwargs)
        return result

    func = memory.cache(func, **cache_kwargs)
    func = wrapper(func)
    func._self_memory = memory  # pyright: ignore[reportAttributeAccessIssue] # noqa: SLF001
    return func


@functools.cache
def new_memory() -> joblib.Memory:
    return joblib.Memory(location=config.joblib.memory.location)


def _filter_keys[KT, VT](mapping: Mapping[KT, VT], keys: Iterable[KT]) -> dict[KT, VT]:
    return toolz.keyfilter(lambda k: k in keys, mapping)



import os
import time
from typing import Literal

import autoregistry

# ruff: noqa: PYI051
# for code-completion
type ClockName = (
    Literal[
        "monotonic",
        "perf",
        "process",
        "thread",
        "time",
        "children-system",
        "children-user",
        "elapsed",
        "system",
        "user",
    ]
    | str
)

CLOCK_REGISTRY = autoregistry.Registry()

CLOCK_REGISTRY["monotonic"] = time.monotonic
CLOCK_REGISTRY["perf"] = time.perf_counter
CLOCK_REGISTRY["process"] = time.process_time
CLOCK_REGISTRY["thread"] = time.thread_time
CLOCK_REGISTRY["time"] = time.time


CLOCK_REGISTRY["children-system"] = lambda: os.times().children_system
CLOCK_REGISTRY["children-user"] = lambda: os.times().children_user
CLOCK_REGISTRY["elapsed"] = lambda: os.times().elapsed
CLOCK_REGISTRY["system"] = lambda: os.times().system
CLOCK_REGISTRY["user"] = lambda: os.times().user


def clock(name: ClockName = "perf") -> float:
    return CLOCK_REGISTRY[name]()



import types
from collections.abc import Iterable
from pathlib import Path

import pydantic

from liblaf.grapes._config._base import BaseModel
from liblaf.grapes._config._paths import paths


class ConfigLoggingTraceback(BaseModel):
    """.

    References:
        1. [`rich.traceback.Traceback.from_exception`](https://rich.readthedocs.io/en/stable/reference/traceback.html#rich.traceback.Traceback.from_exception)
    """

    width: int | None = pydantic.Field(default=None)
    """Number of characters used to traceback."""

    code_width: int | None = pydantic.Field(default=None)
    """Number of code characters used to traceback."""

    extra_lines: int = pydantic.Field(default=3)
    """Additional lines of code to render."""

    show_locals: int = pydantic.Field(default=True)
    """Enable display of local variables."""

    suppress: Iterable[str | types.ModuleType] = pydantic.Field(default_factory=list)
    """Optional sequence of modules or paths to exclude from traceback."""


class ConfigLogging(BaseModel):
    file: Path = pydantic.Field(default=paths.log_file)
    level: str | int = pydantic.Field(default="INFO")
    traceback: ConfigLoggingTraceback = pydantic.Field(
        default_factory=ConfigLoggingTraceback
    )



from . import filters, handlers, profiles, sink
from ._depth_tracker import depth_tracker
from ._init import init
from .filters import CompositeFilter, make_filter
from .handlers import file_handler, jsonl_handler, rich_handler
from .profiles import (
    MixinExceptHook,
    MixinLoguru,
    MixinUnraisableHook,
    Profile,
    ProfileCherries,
    ProfileDefault,
    ProfileLike,
    ProfileName,
    factory,
    ic_arg_to_string_function,
)
from .sink import (
    RichSink,
    RichSinkColumn,
    RichSinkColumnElapsed,
    RichSinkColumnLevel,
    RichSinkColumnLocation,
    RichSinkColumnMessage,
    RichTracebackConfig,
    default_columns,
    default_console,
    default_suppress,
)

__all__ = [
    "CompositeFilter",
    "MixinExceptHook",
    "MixinLoguru",
    "MixinUnraisableHook",
    "Profile",
    "ProfileCherries",
    "ProfileDefault",
    "ProfileLike",
    "ProfileName",
    "RichSink",
    "RichSinkColumn",
    "RichSinkColumnElapsed",
    "RichSinkColumnLevel",
    "RichSinkColumnLocation",
    "RichSinkColumnMessage",
    "RichTracebackConfig",
    "default_columns",
    "default_console",
    "default_suppress",
    "depth_tracker",
    "factory",
    "file_handler",
    "filters",
    "handlers",
    "ic_arg_to_string_function",
    "init",
    "jsonl_handler",
    "make_filter",
    "profiles",
    "rich_handler",
    "sink",
]



from collections.abc import Sequence

import attrs
import loguru

from liblaf.grapes.logging import handlers

from ._default import ProfileDefault
from ._default import default_handlers as _default_handlers


def default_handlers() -> Sequence["loguru.HandlerConfig"]:
    return [*_default_handlers(), handlers.file_handler()]


@attrs.define
class ProfileCherries(ProfileDefault):
    # overrides mixins.LoggingProfileMixinLoguru
    handlers: Sequence["loguru.HandlerConfig"] | None = attrs.field(
        factory=default_handlers
    )



import functools
import types
from collections.abc import Mapping

import loguru

from ._composite import CompositeFilter
from .typed import FilterLike


@functools.singledispatch
def make_filter(f: FilterLike, /) -> FilterLike:
    return f


@make_filter.register(types.NoneType)
def _make_filter_none(_: None, /) -> "loguru.FilterFunction":
    return CompositeFilter()


@make_filter.register(Mapping)
def _make_filter_mapping(by_level: "loguru.FilterDict", /) -> "loguru.FilterFunction":
    return CompositeFilter(by_level=by_level)



import pydantic

from ._base import BaseModel


class ConfigPretty(BaseModel):
    """.

    References:
        1. [wadler_lindig.pformat](https://docs.kidger.site/wadler_lindig/api/#wadler_lindig.pformat)
    """

    width: int | None = pydantic.Field(default=None)
    """a best-effort maximum width to allow. May be exceeded if there are unbroken pieces of text which are wider than this."""

    indent: int = pydantic.Field(default=2)
    """when the contents of a structured type are too large to fit on one line, they will be indented by this amount and placed on separate lines."""

    short_arrays: bool = pydantic.Field(default=True)
    """whether to print a NumPy array / PyTorch tensor / JAX array as a short summary of the form `f32[3,4]` (here indicating a `float32` matrix of shape `(3, 4)`)"""

    hide_defaults: bool = pydantic.Field(default=True)
    """whether to show the default values of dataclass fields."""

    show_type_module: bool = pydantic.Field(default=True)
    """whether to show the name of the module for a type: `somelib.SomeClass` versus `SomeClass`."""

    show_dataclass_module: bool = pydantic.Field(default=False)
    """whether to show the name of the module for a dataclass instance: `somelib.SomeClass()` versus `SomeClass()`."""

    show_function_module: bool = pydantic.Field(default=False)
    """whether to show the name of the module for a function: `<function some_fn>` versus `<function somelib.some_fn>`."""

    respect_pdoc: bool = pydantic.Field(default=True)



from collections.abc import Mapping, Sequence


def deep_merge[KT, VT](
    *mappings: Mapping[KT, VT], append_arrays: bool = False
) -> dict[KT, VT]:
    result: dict[KT, VT] = {}
    for mapping in mappings:
        for key, value in mapping.items():
            if key not in result:
                result[key] = value
            elif isinstance(result[key], Mapping):
                result[key] = deep_merge(  # pyright: ignore[reportArgumentType]
                    result[key],  # pyright: ignore[reportArgumentType]
                    value,  # pyright: ignore[reportArgumentType]
                    append_arrays=append_arrays,
                )
            elif append_arrays and isinstance(result[key], Sequence):
                result[key] = [*result[key], *value]  # pyright: ignore[reportArgumentType, reportGeneralTypeIssues]
    return result



import functools
import os
import sys
from pathlib import Path

import attrs


@attrs.define
class Paths:
    def data(self, path: str | os.PathLike = "", *, mkdir: bool = False) -> Path:
        path = self.working_dir / "data" / path
        if mkdir:
            path.mkdir(parents=True, exist_ok=True)
        return path

    def fig(self, path: str | os.PathLike = "", *, mkdir: bool = False) -> Path:
        path = self.working_dir / "fig" / path
        if mkdir:
            path.mkdir(parents=True, exist_ok=True)
        return path

    @functools.cached_property
    def entrypoint(self) -> Path:
        return Path(sys.argv[0])

    @functools.cached_property
    def log_file(self) -> Path:
        return self.working_dir / "run.log"

    @functools.cached_property
    def working_dir(self) -> Path:
        working_dir: Path = self.entrypoint.resolve().parent
        while working_dir.parts[-1] in {"src"}:
            working_dir = working_dir.parent
        return working_dir


paths: Paths = Paths()



import types
import unittest.mock
from collections.abc import Generator, Sequence

import attrs
import loguru
from rich.console import Console, RenderableType
from rich.text import Text
from rich.traceback import Traceback

from liblaf.grapes import pretty
from liblaf.grapes._config import config

from .columns import (
    RichSinkColumn,
    RichSinkColumnElapsed,
    RichSinkColumnLevel,
    RichSinkColumnLocation,
    RichSinkColumnMessage,
)


def default_columns(*, enable_link: bool = True) -> Sequence[RichSinkColumn]:
    return [
        RichSinkColumnElapsed(),
        RichSinkColumnLevel(),
        RichSinkColumnLocation(enable_link=enable_link),
        RichSinkColumnMessage(),
    ]


def default_console() -> Console:
    return pretty.get_console(stderr=True)


@attrs.define
class RichSink:
    columns: Sequence[RichSinkColumn] = attrs.field(
        converter=attrs.converters.default_if_none(factory=default_columns),
        factory=default_columns,
    )
    console: Console = attrs.field(
        converter=attrs.converters.default_if_none(factory=default_console),
        factory=default_console,
    )

    def __call__(self, message: "loguru.Message", /) -> None:
        record: loguru.Record = message.record
        # TODO: `console.print()` is slow
        self.console.print(
            *self.render(record), overflow="ignore", no_wrap=True, crop=False
        )
        if (excpetion := self.render_exception(record)) is not None:
            self.console.print(excpetion)

    def render(self, record: "loguru.Record", /) -> Generator[RenderableType]:
        for column in self.columns:
            yield column.render(record)

    def render_exception(self, record: "loguru.Record", /) -> RenderableType | None:
        exception: loguru.RecordException | None = record["exception"]
        if exception is None:
            return None
        exc_type: type[BaseException] | None
        exc_value: BaseException | None
        traceback: types.TracebackType | None
        exc_type, exc_value, traceback = exception
        if exc_type is None or exc_value is None:
            return None

        # ? dirty hack to avoid long `repr()` output
        # ref: <https://github.com/Textualize/rich/discussions/3774>
        with unittest.mock.patch("rich.pretty.repr", new=pretty.pformat):
            rich_tb: Traceback = Traceback.from_exception(
                exc_type,
                exc_value,
                traceback,
                **config.logging.traceback.to_dict(exclude_none=False),
            )

        # ? dirty hack to support ANSI in exception messages
        for stack in rich_tb.trace.stacks:
            if pretty.has_ansi(stack.exc_value):
                stack.exc_value = Text.from_ansi(stack.exc_value)  # pyright: ignore[reportAttributeAccessIssue]
        return rich_tb



from collections.abc import Sequence
from typing import override

import attrs
import loguru

from liblaf.grapes.logging import handlers

from . import mixins
from ._abc import Profile


def default_handlers() -> Sequence["loguru.HandlerConfig"]:
    return [handlers.rich_handler()]


def default_levels() -> Sequence["loguru.LevelConfig"]:
    return [{"name": "ICECREAM", "no": 15, "color": "<magenta><bold>", "icon": "🍦"}]


@attrs.define
class ProfileDefault(
    mixins.MixinLoguru,
    mixins.MixinStdlib,
    mixins.MixinIcecream,
    mixins.MixinExceptHook,
    mixins.MixinUnraisableHook,
    Profile,
):
    # overrides mixins.LoggingProfileMixinLoguru
    handlers: Sequence["loguru.HandlerConfig"] | None = attrs.field(
        factory=default_handlers
    )
    levels: Sequence["loguru.LevelConfig"] | None = attrs.field(factory=default_levels)
    level: int | str | None = attrs.field(default=None)

    @override
    def init(self) -> None:
        self.configure_loguru()
        self.clear_stdlib_handlers()
        self.configure_icecream()
        self.configure_excepthook()
        self.configure_unraisablehook()



"""This module provides progress tracking utilities, including a custom rate column for progress bars and a function to track the progress of iterating over sequences."""

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach_stub(__name__, __file__)
del lazy



import contextlib
from collections.abc import Generator

from etils import epy


@contextlib.contextmanager
def optional_imports(package: str, extra: str) -> Generator[None]:
    try:
        yield
    except ImportError as err:
        epy.reraise(
            err, suffix=f"Make sure to install `{package}` with the `{extra}` extra."
        )



import importlib
import types


def try_import(name: str, package: str | None = None) -> types.ModuleType | None:
    try:
        return importlib.import_module(name, package)
    except ImportError:
        return None



from ._cache import MemorizedFunc, cache
from ._wrapt import Decorator, Wrapper, decorator, unbind, unbind_getattr

__all__ = [
    "Decorator",
    "MemorizedFunc",
    "Wrapper",
    "cache",
    "decorator",
    "unbind",
    "unbind_getattr",
]



from typing import override

import attrs
import loguru
from rich.console import RenderableType
from rich.text import Text

from liblaf.grapes import pretty

from ._abc import RichSinkColumn


@attrs.define
class RichSinkColumnLocation(RichSinkColumn):
    enable_link: bool = attrs.field(default=True)

    @override  # impl RichSinkColumn
    def render(self, record: "loguru.Record", /) -> RenderableType:
        location: Text = pretty.rich_location(
            name=record["name"],
            function=record["function"],
            line=record["line"],
            file=record["file"].path,
            enable_link=self.enable_link,
        )
        location.style = "log.path"
        return location



from collections.abc import Iterable
from typing import Literal, override

from rich.console import Console, RenderableType
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    ProgressColumn,
    SpinnerColumn,
    Task,
    TaskID,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)
from rich.progress import Progress as RichProgress
from rich.table import Column
from rich.text import Text

from liblaf.grapes import human, pretty, timing
from liblaf.grapes.logging import depth_tracker


class RateColumn(ProgressColumn):
    """RateColumn is a subclass of ProgressColumn that represents the rate of progress for a given task."""

    unit: str = "it"
    """The unit of measurement for the progress bar."""

    def __init__(self, unit: str = "it", table_column: Column | None = None) -> None:
        """.

        Args:
            unit: The unit of measurement for the progress bar.
            table_column: The table column associated with the progress bar.
        """
        super().__init__(table_column)
        self.unit = unit

    def render(self, task: Task) -> RenderableType:
        """Render the progress speed of a given task.

        Args:
            task: The task for which the speed is to be rendered.

        Returns:
            A text object representing the speed of the task.
        """
        if not task.speed:
            return Text(f"?{self.unit}/s", style="progress.data.speed")
        throughput: str = human.human_throughput(task.speed, self.unit)
        return Text(throughput, style="progress.data.speed")


class Progress(RichProgress):
    timer: timing.Timer | Literal[False]

    def __init__(
        self,
        *columns: str | ProgressColumn,
        console: Console | None = None,
        timer: timing.Timer | Literal[False] | None = None,
    ) -> None:
        if console is None:
            console = pretty.get_console(stderr=True)
        if timer is None:
            timer = timing.timer()
        self.timer = timer
        super().__init__(*columns, console=console)

    @override
    @classmethod
    def get_default_columns(cls) -> tuple[str | ProgressColumn, ...]:  # pyright: ignore[reportIncompatibleMethodOverride]
        return (
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
            MofNCompleteColumn(),
            "[",
            TimeElapsedColumn(),
            "<",
            TimeRemainingColumn(),
            ",",
            RateColumn(),
            "]",
        )

    @override
    @depth_tracker
    def track[T](
        self,
        sequence: Iterable[T],
        total: float | None = None,
        completed: int = 0,
        task_id: TaskID | None = None,
        description: str = "Working...",
        update_period: float = 0.1,
        *,
        timer: timing.Timer | Literal[False] | None = None,
    ) -> Iterable[T]:
        if total is None:
            total = len_safe(sequence)
        if timer := (timer or self.timer):
            sequence = timer(sequence)
            timing.get_timer(sequence).name = description
        with depth_tracker(depth=2):
            yield from super().track(
                sequence,
                total=total,
                completed=completed,
                task_id=task_id,
                description=description,
                update_period=update_period,
            )


def len_safe(iterable: Iterable) -> int | None:
    try:
        return len(iterable)  # pyright: ignore[reportArgumentType]
    except TypeError:
        return None



import logging


class MixinStdlib:
    def clear_stdlib_handlers(self) -> None:
        for logger in logging.root.manager.loggerDict.values():
            if isinstance(logger, logging.PlaceHolder):
                continue
            logger.handlers.clear()
            logger.propagate = True



from . import mixins
from ._abc import Profile
from ._cherries import ProfileCherries
from ._default import ProfileDefault
from ._factory import ProfileLike, ProfileName, factory
from .mixins import (
    MixinExceptHook,
    MixinLoguru,
    MixinUnraisableHook,
    ic_arg_to_string_function,
)

__all__ = [
    "MixinExceptHook",
    "MixinLoguru",
    "MixinUnraisableHook",
    "Profile",
    "ProfileCherries",
    "ProfileDefault",
    "ProfileLike",
    "ProfileName",
    "factory",
    "ic_arg_to_string_function",
    "mixins",
]



from . import columns
from ._sink import RichSink, default_columns, default_console
from ._traceback import RichTracebackConfig, default_suppress
from .columns import (
    RichSinkColumn,
    RichSinkColumnElapsed,
    RichSinkColumnLevel,
    RichSinkColumnLocation,
    RichSinkColumnMessage,
)

__all__ = [
    "RichSink",
    "RichSinkColumn",
    "RichSinkColumnElapsed",
    "RichSinkColumnLevel",
    "RichSinkColumnLocation",
    "RichSinkColumnMessage",
    "RichTracebackConfig",
    "columns",
    "default_columns",
    "default_console",
    "default_suppress",
]



import sys
from typing import Any


def array_kind(obj: Any) -> str | None:
    for module, typename in [
        ("numpy", "ndarray"),
        ("torch", "Tensor"),
        ("jax", "Array"),
        ("mlx.core", "array"),
    ]:
        if module not in sys.modules:
            continue
        typ: type = getattr(sys.modules[module], typename)
        if isinstance(obj, typ):
            return module
    return None



from ._base import BaseConfig, BaseModel
from ._config import Config, config
from ._joblib import ConfigJoblib, ConfigJoblibMemory

__all__ = [
    "BaseConfig",
    "BaseModel",
    "Config",
    "ConfigJoblib",
    "ConfigJoblibMemory",
    "config",
]



from typing import Literal

from ._abc import Profile

# make sure profiles are registered
from ._cherries import ProfileCherries  # noqa: F401
from ._default import ProfileDefault

# for code-completion
type ProfileName = Literal["default", "cherries"] | str  # noqa: PYI051
type ProfileLike = ProfileName | Profile | type[Profile]


def factory(profile: ProfileLike | None = None, /, *args, **kwargs) -> Profile:
    if profile is None:
        return ProfileDefault(*args, **kwargs)
    if isinstance(profile, Profile):
        return profile
    if isinstance(profile, str):
        return Profile[profile](*args, **kwargs)
    return profile(*args, **kwargs)



from collections.abc import Callable
from typing import Any


def clone_param_spec[**P, T](
    _func: Callable[P, T], /
) -> Callable[[Any], Callable[P, T]]:
    def wrapper(wrapped: Any, /) -> Callable[P, T]:
        return wrapped

    return wrapper


def clone_signature[C](_source: C, /) -> Callable[[Any], C]:
    def wrapper(obj: Any) -> C:
        return obj

    return wrapper



import importlib
import types
from collections.abc import Iterable

import pydantic
from environs import env

from liblaf.grapes._config._base import BaseModel


def _show_locals_default() -> bool:
    return env.bool("RICH_TRACEBACK_SHOW_LOCALS", True)


def _suppress_default() -> list[str | types.ModuleType]:
    packages: list[str] = ["liblaf.cherries", "pydantic"]
    suppress: list[str | types.ModuleType] = []
    for package in packages:
        try:
            module: types.ModuleType = importlib.import_module(package)
        except ImportError:
            pass
        else:
            suppress.append(module)
    return suppress


class ConfigLoggingTraceback(BaseModel):
    """.

    References:
        1. [`rich.traceback.Traceback.from_exception`](https://rich.readthedocs.io/en/stable/reference/traceback.html#rich.traceback.Traceback.from_exception)
    """

    width: int | None = pydantic.Field(default=None)
    """Number of characters used to traceback."""

    code_width: int | None = pydantic.Field(default=None)
    """Number of code characters used to traceback."""

    extra_lines: int = pydantic.Field(default=3)
    """Additional lines of code to render."""

    show_locals: int = pydantic.Field(default_factory=_show_locals_default)
    """Enable display of local variables."""

    suppress: Iterable[str | types.ModuleType] = pydantic.Field(
        default_factory=_suppress_default
    )
    """Optional sequence of modules or paths to exclude from traceback."""



import about_time


def human_throughput(value: float, unit: str = "", prec: int | None = None) -> str:
    throughput: about_time.HumanThroughput = about_time.HumanThroughput(value, unit)
    return throughput.as_human(prec)



from collections.abc import Callable, Mapping
from typing import Any, Literal, TypedDict

import pydantic

type EncHook = Callable[[Any], Any]


type IncEx = (
    set[int] | set[str] | Mapping[int, "IncEx | bool"] | Mapping[str, "IncEx | bool"]
)


class PydanticModelDumpOptions(TypedDict, total=False):
    mode: Literal["json", "python"]
    include: IncEx | None
    exclude: IncEx | None
    context: Any | None
    by_alias: bool | None
    exclude_unset: bool
    exclude_defaults: bool
    exclude_none: bool
    round_trip: bool
    warnings: bool | Literal["none", "warn", "error"]
    fallback: Callable[[Any], Any] | None
    serialize_as_any: bool


def enc_hook(
    obj: Any, /, *, pydantic_options: PydanticModelDumpOptions | None = None
) -> Any:
    if isinstance(obj, pydantic.BaseModel):
        pydantic_options = pydantic_options or {}
        return obj.model_dump(**pydantic_options)
    return obj



from ._excepthook import MixinExceptHook
from ._icecream import MixinIcecream, ic_arg_to_string_function
from ._loguru import MixinLoguru
from ._stdlib import MixinStdlib
from ._unraisablehook import MixinUnraisableHook

__all__ = [
    "MixinExceptHook",
    "MixinIcecream",
    "MixinLoguru",
    "MixinStdlib",
    "MixinUnraisableHook",
    "ic_arg_to_string_function",
]



from . import callback
from ._base import BaseTimer
from ._clock import CLOCK_REGISTRY, ClockName, clock
from ._main import timer
from ._timer import Timer
from ._timings import Timings
from ._utils import get_timer

__all__ = [
    "CLOCK_REGISTRY",
    "BaseTimer",
    "ClockName",
    "Timer",
    "Timings",
    "callback",
    "clock",
    "get_timer",
    "timer",
]



import datetime
from typing import override

import loguru
from rich.console import RenderableType
from rich.text import Text

from ._abc import RichSinkColumn


class RichSinkColumnElapsed(RichSinkColumn):
    @override  # impl RichSinkColumn
    def render(self, record: "loguru.Record", /) -> RenderableType:
        elapsed: datetime.timedelta = record["elapsed"]
        hh: int
        mm: int
        ss: int
        mm, ss = divmod(int(elapsed.total_seconds()), 60)
        hh, mm = divmod(mm, 60)
        # TODO: handle longer timedelta
        # ref: <https://docs.pydantic.dev/latest/api/standard_library_types/#datetimetimedelta>
        # format: [[DD]D,]HH:MM:SS[.ffffff]
        # Ex: '1d,01:02:03.000004' or '1D01:02:03.000004' or '01:02:03'
        return Text(
            f"{hh:02d}:{mm:02d}:{ss:02d}.{elapsed.microseconds:06d}", style="log.time"
        )



import enum
from typing import override


class LogLevel(enum.StrEnum):
    """.

    References:
        1. <https://github.com/Delgan/loguru/blob/master/loguru/_defaults.py>
    """

    @override
    @staticmethod
    def _generate_next_value_(
        name: str, start: int, count: int, last_values: list[str]
    ) -> str:
        return name.upper()

    TRACE = enum.auto()
    DEBUG = enum.auto()
    INFO = enum.auto()
    SUCCESS = enum.auto()
    WARNING = enum.auto()
    ERROR = enum.auto()
    CRITICAL = enum.auto()



from collections.abc import Generator, Iterable

import wrapt

from liblaf.grapes.logging import depth_tracker

from ._base import BaseTimer


class TimedIterable[T](wrapt.ObjectProxy):
    __wrapped__: Iterable[T]
    _self_timer: BaseTimer

    def __init__(self, wrapped: Iterable[T], timer: BaseTimer) -> None:
        super().__init__(wrapped)
        if timer.name is None:
            timer.name = "Iterable"
        self._self_timer = timer

    def __iter__(self) -> Generator[T]:
        with depth_tracker():
            self._self_timer.start()
            try:
                for item in self.__wrapped__:
                    self._self_timer.start()
                    yield item
                    self._self_timer.stop()
            finally:
                # When the `for` loop is exhausted, it does not re-enter the loop
                # body. Therefore, the `start()` call after the *last* item is
                # redundant. However, since `timer._start_time` is not used anywhere
                # else, we can safely leave it out.
                self._self_timer.finish()



from pathlib import Path

import platformdirs
import pydantic

from ._base import BaseModel


def _default_joblib_memory_location() -> Path:
    return platformdirs.user_cache_path(appname="joblib")


class ConfigJoblibMemory(BaseModel):
    bytes_limit: int | str | None = pydantic.Field(default="4G")
    location: Path = pydantic.Field(default_factory=_default_joblib_memory_location)


class ConfigJoblib(BaseModel):
    memory: ConfigJoblibMemory = pydantic.Field(default_factory=ConfigJoblibMemory)



from ._base import BaseConfig, BaseModel
from ._config import Config, config
from ._joblib import ConfigJoblib, ConfigJoblibMemory
from ._logging import ConfigLogging, ConfigLoggingTraceback, LogLevel
from ._paths import Paths, paths
from ._pretty import ConfigPretty

__all__ = [
    "BaseConfig",
    "BaseModel",
    "Config",
    "ConfigJoblib",
    "ConfigJoblibMemory",
    "ConfigLogging",
    "ConfigLoggingTraceback",
    "ConfigPretty",
    "LogLevel",
    "Paths",
    "config",
    "paths",
]



import functools
from collections.abc import Hashable
from typing import overload

import attrs
import cytoolz as toolz
import loguru
from loguru import logger

from liblaf.grapes import error


@overload
def normalize_filter_dict(
    value: "loguru.FilterDict", /
) -> dict[str | None, int | bool]: ...
@overload
def normalize_filter_dict(value: None, /) -> None: ...
def normalize_filter_dict(
    value: "loguru.FilterDict | None", /
) -> dict[str | None, int | bool] | None:
    if value is None:
        return None
    return {k: get_level_no(v) for k, v in value.items()}


@functools.singledispatch
def get_level_no(level: str | int | bool) -> int | bool:  # noqa: FBT001
    raise error.DispatchLookupError(get_level_no, (level,))


@get_level_no.register
def _get_level_no_str(level: str) -> int:
    return logger.level(level).no


@get_level_no.register
def _get_level_no_int(level: int) -> int:
    return level


@get_level_no.register
def _get_level_no_bool(level: bool) -> int | bool:  # noqa: FBT001
    return 0 if level else False


def default_filter_by_level() -> dict[str | None, int | bool]:
    return normalize_filter_dict(
        {
            "": "INFO",
            "__main__": "TRACE",
            "liblaf": "DEBUG",
        }
    )


@attrs.define
class CompositeFilter:
    by_level: dict[str | None, int | bool] | None = attrs.field(
        factory=default_filter_by_level, converter=normalize_filter_dict
    )
    by_name: str | None = attrs.field(default=None)
    none: bool = attrs.field(default=False)
    once: bool = attrs.field(default=True)
    inherit: bool = attrs.field(default=True)
    _cache_record_id: int | None = attrs.field(default=None, init=False)
    _cache_result: bool | None = attrs.field(default=None, init=False)
    _once_history: set[Hashable] = attrs.field(factory=set, init=False)

    def __attrs_post_init__(self) -> None:
        if self.inherit:
            self.by_level = toolz.merge(default_filter_by_level(), self.by_level or {})

    def __call__(self, record: "loguru.Record", /) -> bool:
        # When a filter is used by multiple handlers, it will be called multiple
        # times with the same record. We have to cache the result to make
        # `filter_once` work correctly.
        if id(record) == self._cache_record_id:
            assert self._cache_result is not None
            return self._cache_result
        self._cache_record_id = id(record)
        self._cache_result = (
            self.filter_by_level(record)
            and self.filter_by_name(record)
            and self.filter_none(record)
            and self.filter_once(record)
        )
        return self._cache_result

    def filter_by_level(self, record: "loguru.Record") -> bool:
        if self.by_level is None:
            return True
        name: str | None = record["name"]
        while True:
            level: int | bool | None = self.by_level.get(name, None)
            if level is False:
                return False
            if level is not None:
                return record["level"].no >= level
            if not name:
                return True
            index: int = name.rfind(".")
            name = "" if index < 0 else name[:index]

    def filter_by_name(self, record: "loguru.Record") -> bool:
        if self.by_name is None:
            return True
        name: str | None = record["name"]
        if name is None:
            return False
        return f"{name}.".startswith(self.by_name)

    def filter_none(self, record: "loguru.Record") -> bool:
        if not self.none:
            return True
        return record["name"] is not None

    def filter_once(self, record: "loguru.Record") -> bool:
        if not self.once:
            return True
        if not record["extra"].get("once", False):
            return True
        key: Hashable = self._hash_record(record)
        if key in self._once_history:
            return False
        self._once_history.add(key)
        return True

    def set_level(self, module: str | None, level: str | int | bool | None) -> None:  # noqa: FBT001
        if self.by_level is None:
            self.by_level = {}
        if level is None:
            del self.by_level[module]
            return
        self.by_level[module] = get_level_no(level)

    def _hash_record(self, record: "loguru.Record") -> Hashable:
        return (
            record["function"],
            record["level"].no,
            record["line"],
            record["message"],
            record["name"],
        )



import functools
import types
from collections.abc import Callable, Iterable
from typing import Any

from ._timer import Timer


@functools.singledispatch
def _timer_dispatch(*args, **kwargs) -> Any:
    return Timer(*args, **kwargs)


@_timer_dispatch.register(str)
@_timer_dispatch.register(types.NoneType)
def _timer_name(name: str | None, /, **kwargs) -> Timer:
    return Timer(name=name, **kwargs)


@_timer_dispatch.register(Callable)
@_timer_dispatch.register(Iterable)
def _timer_wrapper[T: Callable | Iterable](wrapped: T, /, **kwargs) -> T:
    return Timer(**kwargs)(wrapped)


def timer(*args, **kwargs) -> Any:
    if len(args) > 0:
        return _timer_dispatch(*args, **kwargs)
    return Timer(*args, **kwargs)



from ._ansi import has_ansi
from ._call import pretty_call
from ._console import get_console
from ._func import pretty_func
from ._location import rich_location
from ._wadler_lindig import WadlerLindigOptions, pdoc_attrs, pdoc_custom, pformat

__all__ = [
    "WadlerLindigOptions",
    "get_console",
    "has_ansi",
    "pdoc_attrs",
    "pdoc_custom",
    "pformat",
    "pretty_call",
    "pretty_func",
    "rich_location",
]



import functools
from typing import Any

import attrs
from loguru import logger

from liblaf.grapes import pretty


@attrs.define(slots=False)
class MixinIcecream:
    ic_prefix: str = attrs.field(default="")

    def configure_icecream(self, prefix: str | None = None) -> None:
        try:
            from icecream import ic
        except ImportError:
            return

        if prefix is None:
            prefix = self.ic_prefix

        self.ic_prefix = prefix

        ic.configureOutput(
            prefix=prefix,
            argToStringFunction=ic_arg_to_string_function,
            outputFunction=self.ic_output_function,
        )

    def ic_output_function(self, s: str) -> None:
        logger.opt(depth=2).log("ICECREAM", s)


@functools.singledispatch
def ic_arg_to_string_function(obj: Any) -> str:
    return pretty.pformat(obj)
